// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#ifndef LLOAD_H
#define LLOAD_H

#include "LLoadTypes.h"
#include "LibParse.h"
#include "LibGroup.h"
#include "TLib"
#include "TLibPin"
#include "TLibPGPin"
#include "TLibMVInfo"
#include "NLib"
#include "NCell"

class TLibDelay;
class TLibDelayTable;

class LCell : public TLibCellPG, public TLibMVInfo
{
  __NO_COPYING(LCell);

public:
  typedef UHashMap<UString, BDD>       ModeVals;
  typedef ModeVals::iterator           ModeValIter;
  typedef UHashMap<UString, ModeVals*> ModeDefs;
  typedef ModeDefs::iterator           ModeDefIter;

public:
  LCell(NCell c) : 
    TLibCellPG(),
    d_ncell(c), d_bidirIdx(0), d_inputIdx(0), d_internalIdx(0), d_pinCount(0),
    d_pins(NULL), d_order(NULL),
    d_pinNames(NULL), d_nextStateType(NULL) {}
  ~LCell();

  void createPins(int out, int bidir, int in, int internal);

  int         totalPinCount()        const {return d_pinCount;}
  int         outPinCount()          const {return d_bidirIdx;}
  int         bidirPinCount()        const {return d_inputIdx - d_bidirIdx;}
  int         inPinCount()           const {return d_internalIdx - d_inputIdx;}
  int         internalPinCount()     const {return d_pinCount - d_internalIdx;}
  int         externalPinCount()     const {return d_internalIdx;}
  UAggSlice   allPins()              const {return UAggSlice(this, 0, totalPinCount());}
  UAggSlice   outPins()              const {return UAggSlice(this, 0, outPinCount());}
  UAggSlice   bidirPins()            const {return UAggSlice(this, d_bidirIdx, bidirPinCount());}
  UAggSlice   inPins()               const {return UAggSlice(this, d_inputIdx, inPinCount());}
  UAggSlice   internalPins()         const {return UAggSlice(this, d_internalIdx, internalPinCount());}
  UAggSlice   externalPins()         const {return UAggSlice(this, 0, d_internalIdx);}
  int         outPin()               const {uassert(d_bidirIdx==1); return 0;}
  int         inPin()                const {uassert(d_pinCount-d_inputIdx==1); return d_inputIdx;}
  // the foll two include bidirs
  UAggSlice   driverPins()           const {return UAggSlice(this, 0, d_inputIdx);}
  UAggSlice   loadPins()             const {return UAggSlice(this, d_bidirIdx, d_pinCount-d_bidirIdx);}
  TLibPin&    pin(const UAggIter& i) const;
  TLibPin&    pin(int i)             const;
  int         pinOrder(int i)        const {return d_order[i];}
  int*        pinOrder()             const {return d_order;}
  void        swapPins(int i, int j);
  void        permutePins(UAggSlice s, int permute[], int idxOffset);
  void        setNextStateType(int i, const UString& nm)
                                           {d_nextStateType[i] = nm;}
  const UString&  nextStateType(int i)
                                     const {return d_nextStateType[i];}
  void        setPinName(int i, const UString& nm) {d_pinNames[i] = nm;}
  const UString&  pinName(int i)     const {return d_pinNames[i];}
  LLoadBundleMap& bundles()                {return d_bundles;}
  LLoadBusMap&    busses()                 {return d_busses;}
  LLoadPinMap&    pinMap()                 {return d_pinMap;}
  NCell       ncell()                const {return d_ncell;}

  UAggType*       findType(const UString& nm) const;
  void            addType(const UString& nm, UAggType* t);
  void            quickAddType(const UString& nm, UAggType* t);

  BDD         findModeValue(const UString& mode, const UString& modeVal, const UTrace& t);
  bool        addModeValue(ModeVals* modeVals, const UString& modeVal, BDD f);
  ModeVals*   addMode(const UString& mode, const UTrace& t);

private:
  NCell       d_ncell;
  int         d_bidirIdx;
  int         d_inputIdx;
  int         d_internalIdx;
  int         d_pinCount;
  TLibPin*    d_pins;
  int*        d_order;
  UString*    d_pinNames;
  UString*    d_nextStateType;
  LLoadBundleMap d_bundles;
  LLoadBusMap    d_busses;
  LLoadPinMap    d_pinMap;
  TLibTypeHashMap d_types;
  ModeDefs        d_modeDefs;
};

inline void
LCell::createPins(int out, int bidir, int in, int internal)
{
  d_bidirIdx = out;
  d_inputIdx = d_bidirIdx + bidir;
  d_internalIdx = d_inputIdx + in;
  d_pinCount = d_internalIdx + internal;
  d_pins = new IN_CURRENT_POOL TLibPin[d_pinCount];
  d_order = new IN_CURRENT_POOL int[d_pinCount];
  d_pinNames = new IN_CURRENT_POOL UString[d_pinCount];
  d_nextStateType = new IN_CURRENT_POOL UString[d_pinCount];
  for (int i = 0; i < d_pinCount; i++) {
    d_order[i] = i;
  }
}

inline TLibPin&
LCell::pin(const UAggIter& i) const
{
  uassert(!i.owner() || i.owner() == this);
  return d_pins[i.current()];
}

inline TLibPin&
LCell::pin(int i) const
{
  return d_pins[i];
}

inline void
LCell::swapPins(int i, int j)
{
  // still need to handle busses/bundles
  TLibPin tmp;
  tmp.grab(d_pins[i]);
  d_pins[i].grab(d_pins[j]);
  d_pins[j].grab(tmp);

  int tmpI;
  tmpI = d_order[i];
  d_order[i] = d_order[j];
  d_order[j] = tmpI;
}

inline UAggType*
LCell::findType(const UString& nm) const
{
  TLibTypeHashMapIter iter = d_types.find(nm);
  return (iter ? iter.value() : NULL);
}

inline void
LCell::quickAddType(const UString& nm, UAggType* t)
{
  uassert(!findType(nm));
  d_types[nm] = t;
}

inline void
LCell::addType(const UString& nm, UAggType* t)
{
  UAggType* oldT = findType(nm);
  if (oldT) {
    uassert (oldT == t);
  } else {
    d_types[nm] = t;
  }
}
  
// ----------------------------------------------------------------------------
// class LLoad - grouping of APIs used to extract library DB from liberty
//               parse tree
// ----------------------------------------------------------------------------
class LLoad 
{
public:
  LLoad(const char* filename, NDb db, bool writeOut, bool printStats,
        bool warn);
  NLib library() {return d_lib;}

private:
  bool warnNoSupport()                        { return d_warnNoSupport; }
  void parseTree2Db(NDb db);
  void addLib(const LibGroup* llib, NDb db);
  void processLibAttr(const LibGroup* llib);
  void processLibSubgroup(const LibGroup* llib, NDb db);
  void addLutTemplate(const LibGroup* ltemp);
  void addCell(const LibGroup* lcell);
  void preProcessCellAttr(const LibGroup* lcell, bool &isMvCell);
  void processCellAttr(const LibGroup* lcell, NCell cell,
		       LLoadRailList& railList);
  void processRailConnAttrs(LCell& lcell, NCell cell,
			    const LLoadRailList& railList);
  void preProcessCellSubgroup(const LibGroup* lcell, bool& isMvCell);
  void processCellSubgroup(const LibGroup* lcell, NCell cell);
  void instantiateCellPins(const LibGroup* lcell, NCell cell,
			   const LLoadRailList& railList);
  void processSeqInfo(const LibGroup* seq, TLibInternalPinC* tpin,
		      uint_t seqType, NCell cell, int outCount,
		      LLoadPinMap& pinMap);
  void countPins(const LibGroup* lcell, LCell& tmpCell, 
		 int& output, int& bidir, int& input, int& internal, 
		 int& pgOutput, int& pgBidir, int& pgInput, int& pgInternal,
		 LibGroup*& testCell, LibGroup*& coreCell);
  void getPinMap(const LibGroup* lcell, const LibGroup* scanCore,
		 LCell& tmpCell);
  void processInternalSeqPins(const LibGroup* seq, bool isTest, UAggIter& intPinIter,
			      LCell& tmpCell);
  void processPGPins(const LibGroup* cell,
		     LCell&);
  void processPins(const LibGroup* cell, const LibGroup* testCell,
		   LCell&);
  void processPGPinAttr(const LibGroup* lpin, TLibPGPinC* tpin,
			TLibDir dir,
			LCell& lcell, int pinIdx);
  void processPinDetails(const LibGroup* cell, const LibGroup* testCell,
			 LCell&);
  void createPins(const LibGroup* lcell, const LibGroup* scanCore,
		  LCell& tmpCell);
  void createPGPins(LCell& tmpCell);
//   void processPin(const LibGroup* lpin, TLibExternalPinC* tpin, TLibDir dir,
// 		  int outCount, LLoadPinMap& pinMap);
  void processTestPinAttr(const LibGroup* lpin, TLibExternalPinC* tpin,
			  int outCount, LLoadPinMap& pinMap, LLoadBusMap* busses,
			  UAggType* busType, int grpIdx);
  void processPinAttr(const LibGroup* lpin, TLibExternalPinC* tpin, TLibDir dir,
		      LCell& lcell, int pinIdx,
		      int outCount, LLoadPinMap& pinMap, LLoadBundleMap* bundles,
		      LLoadBusMap* busses, UAggType* busType, int grpIdx);
  void processBundlePinAttributes(const LibGroup* bundle, LCell& tmpCell,
				  UAggIter& outPinIter, UAggIter& bidirPinIter,
				  UAggIter& inPinIter, UAggIter& intPinIter);
  void processBundlePinSubgroup(const LibGroup* bundle, LCell& tmpCell, uint_t& delayCount);
  void processBusPinAttributes(const LibGroup* bus, LCell& tmpCell,
			       UAggIter& outPinIter, UAggIter& bidirPinIter,
			       UAggIter& inPinIter, UAggIter& intPinIter,
			       const LibGroup* testCell);
  void processBusPinSubgroup(const LibGroup* bus, LCell& tmpCell, uint_t& delayCount);
  void processPinSubgroup(const LibGroup*, int toIdx,
			  LCell&, LLoadPinMap&, LLoadBundleMap* bundles,
			  LLoadBusMap* busses, UAggType* busType,
			  int grpIdx, uint& delayCount, bool dontAddIfExist);
  void processRelatedPin(TLibDelay* orig, const char* relatedPinStr,
			 TLibTimingPinC* tpin, int toIdx, LCell&,
			 LLoadPinMap& pinMap, LLoadBundleMap* bundles,
			 LLoadBusMap* busses, UAggType* busType,
			 int grpIdx, uint_t& delayCount, bool dontAddIfExist);
  void processTimingGroup(LibGroup*, TLibTimingPinC*, int toIdx, int outCount,
			  LCell&, LLoadPinMap&, LLoadBundleMap* bundles,
			  LLoadBusMap* busses, UAggType* busType,
			  int grpIdx, uint& delayCount, bool dontAddIfExist);
  void processTimingAttr(LCell& cell, LibGroup* ltiming, TLibDelay* d, int outCount,
			 LLoadPinMap& pinMap, LLoadBundleMap* bundles,
			 LLoadBusMap* busses, UAggType* busType,
			 int grpIdx, const char*& relatedPin,
			 const char*& relatedOutPin);
  void processTimingSubgroup(LibGroup* ltiming, TLibDelay* d);
  void processTimingTable(LibGroup* ltable, TLibDelayTable* ttable);
  void processValueList(LibValEList* valList, TLibDelayTable* ttable);
  void processScalingFactorsGroup(const LibGroup* grp);
  void processOperatingCndsGroup(const LibGroup* grp);
  void processPowerSupplyGroup(const LibGroup* psGroup);
  void processDefaultArcs(TLibTimingPinC*, uint_t);
  int  processClkGateAttr(const char* attr, TEdge&, NClkGtRegType&, NClkGtTestType&);

  void processCellTypes(const LibGroup*, LCell&);
  void processGeneratedClocks(const LibGroup*, LCell&);
  void createTypeForLib(const LibGroup*);
  UAggType* processTypeGroup(const LibGroup*);

  UAggType* findType(const UString& nm, LCell& lcell);

  void processModeDefinitions(const LibGroup*, LCell&);

private:
  NLib        d_lib;
  TLibCorner  d_libCorner;
  bool        d_warnNoSupport : 1;
};

#endif

