
/******************************************************************************
    Copyright (c) 1996-2000 Synopsys, Inc.    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/tapinprogram.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

/* declarations for structures to store the contents of liberty files */
#ifndef LIBGROUP_H
#define LIBGROUP_H

#include "LibParse.h"
#include "LibDefine.h"
#include "LibAttr.h"
#include "LibError.h"
#include "UMem"

class LibGroup;
class LibTypedName;

typedef UHashList<UString>                  LibNameHashList;
typedef UEList<LibAttr>                     LibAttrEList;
typedef UList<LibAttr*>                     LibAttrPList;
typedef UHashMap<UString, LibAttrPList>     LibAttrHash;
typedef UEHash<LibDefine, UString>          LibDefineEHash;
typedef UEList<LibGroup>                    LibGroupEList;
typedef UHashMap<LibTypedName, LibGroup*>   LibGroupHash;

typedef LibGroupEList::iterator          LibGroupIter;
typedef LibAttrEList::iterator           LibAttrIter;
typedef LibAttrPList::iterator           LibAttrPIter; 
typedef LibDefineEHash::iterator         LibDefineIter;
typedef LibNameHashList::iterator        LibNameIter;

#if 0
class LibGroupIter : public LibGroupEList::iterator
{
public:
  LibGroup* next() {
    if ((operator const bool())) {
      return operator++(0);
    } else {
      return NULL;
    }
  }
};
#endif
    
class LibTypedName {
public:
  LibTypedName() {}
  LibTypedName(const UString& t, const UString& n) : d_type(t), d_name(n) {}

  void name(UString& n) { d_name = n; }
  void type(UString& t) { d_type = t; }
  const UString& name() const { return d_name; }
  const UString& type() const { return d_type; }

private:
  UString d_type;
  UString d_name;
};

inline static uint_t
LibTypedNameHashFunc(const LibTypedName& t_n)
{
  uint_t h = 0;
  const char *k;

  for (k=t_n.type(); *k; ++k) {
    h = 13 * h + *k;
  }
  for (k=t_n.name(); *k; ++k) {
    h = 13 * h + *k;
  }
  return h;
}

template <> struct UEHashFunc<LibTypedName> {
  uint_t operator()(const LibTypedName& t_n) const {
    return LibTypedNameHashFunc(t_n); }
};

inline static bool
LibTypedNameHashEq(const LibTypedName& t1, const LibTypedName& t2)
{
  return (t1.type() == t2.type() && t1.name() == t2.name());
}

template <> struct UEHashEq<LibTypedName> {
  uint_t operator()(const LibTypedName& t1, const LibTypedName& t2) const {
    return LibTypedNameHashEq(t1, t2); }
};

class LibGroup : public UEListNode
{
public:

  LibGroup() : d_typeID(0), d_owner(NULL) {}

  LibGroup(const UString& nm, const UString& t, uint_t ID,
	   LibGroup *parent, const LibFileInfo& f);

  ~LibGroup();

  const UString&     type()                   const { return d_type; }
  LibGroup*          owner()                        { return d_owner; }
  int                lineno()                 const { return d_info.lineno(); }
  const char*        filename()               const { return LibParse::filename(); }

  void               fileInfo(const LibFileInfo& f) { d_info = f; }
  uint_t             typeID()                 const { return d_typeID; }

  bool               moreThanOneName()        const {
    return (d_nameHashList.begin() != d_nameHashList.end());
  }
  LibAttr*           createAttr(const UString& nm, LibAttrType type,
				const LibFileInfo& f, LibErrorT& err);
  LibDefine*         createDefine(const UString& nm,
				  const UString& allowed_group_name,
				  LibValType valtype, const LibFileInfo& f,
				  LibErrorT& err);
  LibGroup*          createGroup(const UString& nm, const UString& typ,
			 const LibFileInfo& f, LibErrorT& err);

  void               addName(const UString& nm, LibErrorT& err);
  void               deleteName(const UString& nm, LibErrorT& err);

  LibGroup*          findGroup(const UString& nm, const UString& typ,
			       LibErrorT& err) const ;

  LibAttrPIter       findAttr(const UString& nm, LibErrorT& err) const ;
  LibDefine*         findDefine(const UString& nm, LibErrorT& err) const ;

  inline void        removeAttr(LibAttr*);
  inline void        removeDefine(LibDefine*);
  void               removeGroup(LibGroup*);

  LibGroupIter       groupIter()      const { return d_groupEList.begin(); }
  LibAttrIter        attrIter()       const { return d_attrEList.begin(); }
  LibDefineIter      defineIter()     const { return d_defineEHash.beginRandom(); }
  LibNameIter        nameIter()       const { return d_nameHashList.begin(); }
  uint_t             numberOfNames()  const { return d_nameHashList.size(); }

  LibGroup*          firstGroup()     const;
  LibAttr*           firstAttr()      const;
  const UString&     firstName()      const;

  void               moveBefore(LibGroup* targetGroup);
  void               moveAfter(LibGroup* targetGroup);

  void               dump(FILE* f, const char* indent);
  void               createFloatingDefineForCellArea(const char* string);

  LibGroup*          getThis()              { return this; }

private:

  LibGroup*          createTopLevelGroup(const UString& nm, const UString& typ,
					 const LibFileInfo& f, LibErrorT& err);
  LibGroup*          createGroupInGroup(const UString& nm, const UString& typ,
					const LibFileInfo& f, LibErrorT& err);
  int                dontAddHash(uint_t grpType);

private:

  LibNameHashList    d_nameHashList;
  UString            d_type;
  uint_t             d_typeID;

  LibAttrEList       d_attrEList;
  LibAttrHash        d_attrHash;
  LibDefineEHash     d_defineEHash;
  LibGroupEList      d_groupEList;
  LibGroupHash       d_groupHash;

  LibGroup*          d_owner;
  LibFileInfo        d_info;

};
    
inline void
LibGroup::removeAttr(LibAttr* attr)
{
  d_attrHash.remove(attr->name());
  d_attrEList.remove(attr);

}

inline void
LibGroup::removeDefine(LibDefine* def)
{
  d_defineEHash.remove(def->name());
}

inline LibGroup*
LibGroup::firstGroup() const {
  return groupIter();
}

inline LibAttr*
LibGroup::firstAttr() const {
  return attrIter();
}

inline const UString&
LibGroup::firstName() const {
  return *nameIter();
}

#endif

